use crate::{
    comments::{CommentNode, ItemWithAuthor},
    nav::NavState,
    profiles::Profile,
    Error, OptionExt, State,
};
use actix_web::{web, HttpRequest, HttpResponse, Scope};
use chrono::{DateTime, Utc};
use hyaenidae_profiles::store::{File, FileStore, SubmissionStore};
use hyaenidae_toolkit::{Button, FileInput, TextInput};
use std::collections::HashMap;
use uuid::Uuid;

mod middleware;

pub(super) fn scope() -> Scope {
    web::scope("/submissions")
        .service(
            web::resource("/create")
                .route(web::get().to(files_page))
                .route(web::post().to(upload_files)),
        )
        .service(
            web::scope("/{submission_id}")
                .route("", web::get().to(submission_page))
                .service(
                    web::resource("/report")
                        .route(web::get().to(report_page))
                        .route(web::post().to(report)),
                )
                .route("/report-success", web::get().to(report_success_page))
                .service(
                    web::resource("/comment")
                        .route(web::get().to(route_to_update_page))
                        .route(web::post().to(create_comment)),
                )
                .service(
                    web::resource("/update")
                        .route(web::get().to(update_page))
                        .route(web::post().to(update_submission)),
                )
                .service(
                    web::resource("/add-file")
                        .route(web::get().to(route_to_update_page))
                        .route(web::post().to(add_file)),
                )
                .service(
                    web::resource("/remove-file")
                        .route(web::get().to(route_to_update_page))
                        .route(web::post().to(remove_file)),
                )
                .service(
                    web::resource("/publish")
                        .route(web::get().to(route_to_update_page))
                        .route(web::post().to(publish_submission)),
                )
                .service(
                    web::resource("/delete")
                        .route(web::get().to(route_to_update_page))
                        .route(web::post().to(delete_submission)),
                ),
        )
}

pub struct ReportView {
    pub(crate) submission: Submission,
    pub(crate) profile: Profile,
    pub(crate) input: TextInput,
}

impl ReportView {
    async fn build(submission_id: Uuid, dark: bool, state: &State) -> Result<Self, Error> {
        let submission_store = state.profiles.store.submissions.clone();
        let file_store = state.profiles.store.files.clone();
        let submission = web::block(move || {
            Submission::from_stores(submission_id, &submission_store, &file_store)
        })
        .await?;

        let profile_id = submission.profile_id();
        let profile_store = state.profiles.store.profiles.clone();
        let file_store = state.profiles.store.files.clone();
        let profile =
            web::block(move || Profile::from_stores(profile_id, &profile_store, &file_store))
                .await?;

        let mut input = TextInput::new("body");
        input
            .title("Report")
            .placeholder("Say why you are reporting this submission")
            .textarea()
            .dark(dark);

        Ok(ReportView {
            submission,
            profile,
            input,
        })
    }

    pub(crate) fn tiles(&self) -> Vec<crate::profiles::SubmissionView> {
        self.submission
            .files()
            .iter()
            .map(|f| {
                crate::profiles::SubmissionView::from_parts(
                    &self.submission.inner,
                    &self.profile,
                    f,
                )
            })
            .collect()
    }
}

async fn report_page(
    submission_id: web::Path<Uuid>,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    let view = ReportView::build(submission_id.into_inner(), nav_state.dark(), &state).await?;

    if let Some(res) = can_view(Some(&profile), &view.profile, &view.submission, &state).await? {
        return Ok(res);
    }

    crate::rendered(HttpResponse::Ok(), |cursor| {
        crate::templates::submissions::report(cursor, &view, &nav_state)
    })
}

#[derive(Clone, Debug, serde::Deserialize)]
struct ReportForm {
    body: String,
}

const MAX_REPORT_LEN: usize = 1000;

async fn report(
    form: web::Form<ReportForm>,
    submission_id: web::Path<Uuid>,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    let mut view = ReportView::build(submission_id.into_inner(), nav_state.dark(), &state).await?;

    if let Some(res) = can_view(Some(&profile), &view.profile, &view.submission, &state).await? {
        return Ok(res);
    }

    let form = form.into_inner();

    let error = if form.body.len() > MAX_REPORT_LEN {
        format!("Must be shorter than {} characters", MAX_REPORT_LEN)
    } else if form.body.trim().is_empty() {
        format!("Must be present")
    } else {
        use hyaenidae_profiles::apub::actions::CreateReport;
        let res = state
            .profiles
            .run(CreateReport::from_submission(
                view.submission.id(),
                view.profile.id(),
                Some(form.body.clone()),
            ))
            .await;

        match res {
            Ok(_) => return Ok(to_report_success_page(view.submission.id())),
            Err(e) => e.to_string(),
        }
    };

    view.input.error_opt(Some(error)).value(&form.body);

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::report(cursor, &view, &nav_state)
    })
}

async fn report_success_page(
    submission_id: web::Path<Uuid>,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    let view = ReportView::build(submission_id.into_inner(), nav_state.dark(), &state).await?;

    if let Some(res) = can_view(Some(&profile), &view.profile, &view.submission, &state).await? {
        return Ok(res);
    }

    crate::rendered(HttpResponse::Ok(), |cursor| {
        crate::templates::submissions::report_success(cursor, &view, &nav_state)
    })
}

fn to_report_success_page(submission_id: Uuid) -> HttpResponse {
    crate::redirect(&format!("/submissions/{}/report-success", submission_id))
}

pub struct SubmissionState {
    pub(crate) submission: Submission,
    pub(crate) title: TextInput,
    pub(crate) description: TextInput,
    pub(crate) file: FileInput,
    pub(crate) file_error: Option<String>,
    pub(crate) remove_file_error: HashMap<Uuid, String>,
    pub(crate) publish_error: Option<String>,
    pub(crate) delete_error: Option<String>,
}

impl SubmissionState {
    fn new(submission: Submission, dark: bool) -> Self {
        let mut title = title_input(dark);
        title.value(submission.inner.title());

        let mut description = description_input(dark);
        if let Some(text) = submission.inner.description() {
            description.value(text);
        }

        SubmissionState {
            submission,
            title,
            description,
            file: file_input(dark),
            file_error: None,
            remove_file_error: HashMap::new(),
            publish_error: None,
            delete_error: None,
        }
    }

    pub(crate) fn is_published(&self) -> bool {
        self.submission.inner.published().is_some()
    }

    pub(crate) fn view_path(&self) -> String {
        self.submission.view_path()
    }

    pub(crate) fn update_path(&self) -> String {
        format!("/submissions/{}/update", self.submission.id())
    }

    pub(crate) fn add_file_path(&self) -> String {
        format!("/submissions/{}/add-file", self.submission.id())
    }

    pub(crate) fn remove_file_path(&self) -> String {
        format!("/submissions/{}/remove", self.submission.id())
    }

    pub(crate) fn publish_path(&self) -> String {
        format!("/submissions/{}/publish", self.submission.id())
    }

    pub(crate) fn delete_path(&self) -> String {
        format!("/submissions/{}/delete", self.submission.id())
    }

    pub(crate) fn files(&self) -> &[File] {
        &self.submission.files
    }

    fn title(&mut self, text: &str) -> &mut Self {
        self.title.value(text);
        self
    }

    fn title_error(&mut self, error: String) -> &mut Self {
        self.title.error_opt(Some(error));
        self
    }

    fn description(&mut self, text: &str) -> &mut Self {
        self.description.value(text);
        self
    }

    fn description_error(&mut self, error: Option<String>) -> &mut Self {
        self.description.error_opt(error);
        self
    }

    fn file_error(&mut self, error: String) -> &mut Self {
        self.file_error = Some(error);
        self
    }

    fn remove_file_error(&mut self, file_id: Uuid, error: String) -> &mut Self {
        self.remove_file_error.insert(file_id, error);
        self
    }

    fn publish_error(&mut self, error: String) -> &mut Self {
        self.publish_error = Some(error);
        self
    }

    fn delete_error(&mut self, error: String) -> &mut Self {
        self.delete_error = Some(error);
        self
    }
}

#[derive(Clone, Debug)]
pub struct Submission {
    pub(crate) inner: hyaenidae_profiles::store::Submission,
    files: Vec<File>,
}

impl Submission {
    pub(crate) fn id(&self) -> Uuid {
        self.inner.id()
    }

    pub(crate) fn view_path(&self) -> String {
        format!("/submissions/{}", self.id())
    }

    pub(crate) fn report_path(&self) -> String {
        format!("/submissions/{}/report", self.id())
    }

    pub(crate) fn profile_id(&self) -> Uuid {
        self.inner.profile_id()
    }

    fn has_title(&self) -> bool {
        !self.inner.title().trim().is_empty()
    }

    pub(crate) fn title(&self) -> String {
        if self.published().is_none() {
            if !self.has_title() {
                "Not Published".to_owned()
            } else {
                format!("Not Published: {}", self.inner.title())
            }
        } else {
            self.inner.title().to_owned()
        }
    }

    pub(crate) fn description(&self) -> Option<&str> {
        if let Some(description) = self.inner.description() {
            if description.len() == 0 {
                None
            } else {
                Some(description)
            }
        } else {
            None
        }
    }

    pub(crate) fn files(&self) -> &[File] {
        &self.files
    }

    pub(crate) fn published(&self) -> Option<DateTime<Utc>> {
        self.inner.published()
    }

    pub(crate) fn from_stores(
        submission_id: Uuid,
        submission_store: &SubmissionStore,
        file_store: &FileStore,
    ) -> Result<Self, Error> {
        let inner = submission_store.by_id(submission_id)?.req()?;

        let mut files = vec![];
        for file in inner.files() {
            files.push(file_store.by_id(*file)?.req()?);
        }

        Ok(Submission { inner, files })
    }

    pub(crate) fn from_id(submission_id: Uuid, state: &State) -> Result<Self, Error> {
        Self::from_stores(
            submission_id,
            &state.profiles.store.submissions,
            &state.profiles.store.files,
        )
    }
}

async fn files_page(_: Profile, nav_state: NavState) -> Result<HttpResponse, Error> {
    let file_input = file_input(nav_state.dark());

    crate::rendered(HttpResponse::Ok(), |cursor| {
        crate::templates::submissions::create(cursor, &file_input, None, &nav_state)
    })
}

async fn upload_files(
    profile: Profile,
    req: HttpRequest,
    pl: web::Payload,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    let res = state.profiles.upload_image(req, pl.into_inner()).await;

    let error = match res {
        Ok(file_ids) if file_ids.len() == 1 => {
            use hyaenidae_profiles::apub::actions::CreateSubmission;

            let res = state
                .profiles
                .run(CreateSubmission::from_file(profile.id(), file_ids[0]))
                .await;

            match res {
                Ok(Some(submission_id)) => return Ok(to_update_page(submission_id)),
                Ok(None) => "Error creating submission".to_string(),
                Err(e) => e.to_string(),
            }
        }
        Ok(_) => "Incorrect number of files".to_owned(),
        Err(e) => e.to_string(),
    };

    let file_input = file_input(nav_state.dark());

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::create(cursor, &file_input, Some(error), &nav_state)
    })
}

#[derive(Clone, Debug, serde::Deserialize)]
struct FileQuery {
    file_page: usize,
}

pub struct SubmissionView {
    pub(crate) profile: Option<Profile>,
    pub(crate) submission: Submission,
    pub(crate) poster: Profile,
    pub(crate) current_file: File,
    pub(crate) file_num: usize,
    nav: Vec<Button>,
    pub(crate) is_self: bool,
    pub(crate) comment_input: TextInput,
    pub(crate) comments: CommentNode,
}

impl SubmissionView {
    pub(crate) fn nav(&self) -> Vec<&Button> {
        self.nav.iter().collect::<Vec<_>>()
    }

    pub(crate) fn update_path(&self) -> String {
        format!("/submissions/{}/update", self.submission.id())
    }

    pub(crate) fn comment_path(&self) -> String {
        format!("/submissions/{}/comment", self.submission.id())
    }

    pub(crate) fn replying_to(&self) -> ItemWithAuthor {
        ItemWithAuthor::from_submission(self.submission.clone(), self.poster.clone())
    }
}

fn comment_input(dark: bool) -> TextInput {
    let mut input = TextInput::new("body");
    input
        .title("Comment")
        .placeholder("Comment on this submission")
        .textarea()
        .dark(dark);
    input
}

async fn can_view(
    profile: Option<&Profile>,
    poster: &Profile,
    submission: &Submission,
    state: &State,
) -> Result<Option<HttpResponse>, Error> {
    if poster.login_required() && profile.is_none() {
        return Ok(Some(crate::to_404()));
    }

    let is_self = profile
        .as_ref()
        .map(|p| p.id() == poster.id())
        .unwrap_or(false);

    if submission.inner.published().is_none() && !is_self {
        return Ok(Some(crate::to_404()));
    }

    let is_follower_or_self = if let Some(profile) = profile.as_ref() {
        let is_follower = state
            .profiles
            .store
            .view
            .follows
            .by_forward(poster.id(), profile.id())?
            .is_some();

        is_follower || is_self
    } else {
        false
    };

    if submission.inner.is_followers_only() && !is_follower_or_self {
        return Ok(Some(crate::to_404()));
    }

    Ok(None)
}

fn submission_nav(
    page: Option<FileQuery>,
    submission: &Submission,
    nav_state: &NavState,
) -> (Vec<Button>, File, usize) {
    let file_count = submission.files().len();
    let file_num: usize = if let Some(page) = page {
        file_count.saturating_sub(1).min(page.file_page)
    } else {
        0
    };

    let current_file = submission.files()[file_num].clone();

    let mut nav = vec![];
    if file_num > 0 {
        let button = Button::secondary("Previous");
        button
            .href(&format!(
                "/submissions/{}?file_page={}",
                submission.id(),
                file_num.saturating_sub(1)
            ))
            .dark(nav_state.dark());

        nav.push(button);
    }
    if let Some(pictrs) = current_file.pictrs() {
        let button = Button::secondary("Download");
        button
            .href(&crate::images::largest_image(
                pictrs.key(),
                crate::images::ImageType::Png,
            ))
            .dark(nav_state.dark());

        nav.push(button);
    }
    let btn = Button::primary_outline("Report");
    btn.href(&submission.report_path()).dark(nav_state.dark());
    nav.push(btn);
    if file_num < file_count.saturating_sub(1) {
        let button = Button::secondary("Next");
        button
            .href(&format!(
                "/submissions/{}?file_page={}",
                submission.id(),
                file_num.saturating_add(1)
            ))
            .dark(nav_state.dark());

        nav.push(button);
    }

    (nav, current_file, file_num)
}

async fn submission_page(
    profile: Option<Profile>,
    submission: Submission,
    page: Option<web::Query<FileQuery>>,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    let poster = Profile::from_id(submission.profile_id(), &state)?;

    let viewer = profile.as_ref().map(|p| p.id());
    let is_self = viewer.map(|id| id == poster.id()).unwrap_or(false);

    if let Some(res) = can_view(profile.as_ref(), &poster, &submission, &state).await? {
        return Ok(res);
    }

    let (nav, current_file, file_num) =
        submission_nav(page.map(|p| p.into_inner()), &submission, &nav_state);

    let comments = CommentNode::from_submission(submission.id(), viewer, &state)?;

    let view = SubmissionView {
        profile,
        submission,
        poster,
        current_file,
        file_num,
        nav,
        is_self,
        comment_input: comment_input(nav_state.dark()),
        comments,
    };

    crate::rendered(HttpResponse::Ok(), |cursor| {
        crate::templates::submissions::public(cursor, &view, &nav_state)
    })
}

async fn update_page(
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
) -> Result<HttpResponse, Error> {
    if submission.profile_id() != profile.id() {
        return Ok(crate::to_404());
    }

    let state = SubmissionState::new(submission, nav_state.dark());

    crate::rendered(HttpResponse::Ok(), |cursor| {
        crate::templates::submissions::update(cursor, &state, &nav_state)
    })
}

const MAX_TITLE_LEN: usize = 50;
const MAX_DESCRIPTION_LEN: usize = 500;
const MAX_COMMENT_LEN: usize = 500;

fn validate_title(title: &str) -> Option<String> {
    if title.len() > MAX_TITLE_LEN {
        return Some(format!("Must be shorter than {} characters", MAX_TITLE_LEN));
    }

    if title.trim().len() == 0 {
        return Some(format!("Must be present"));
    }

    None
}

fn validate_description(description: &str) -> Option<String> {
    if description.len() > MAX_DESCRIPTION_LEN {
        return Some(format!(
            "Must be shorter than {} characters",
            MAX_DESCRIPTION_LEN
        ));
    }

    None
}

#[derive(Clone, Debug, serde::Deserialize)]
struct CommentForm {
    body: String,
}

async fn create_comment(
    form: web::Form<CommentForm>,
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
    page: Option<web::Query<FileQuery>>,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    use hyaenidae_profiles::apub::actions::CreateComment;
    let form = form.into_inner();

    let poster = Profile::from_id(submission.profile_id(), &state)?;

    if let Some(res) = can_view(Some(&profile), &poster, &submission, &state).await? {
        return Ok(res);
    }

    let error = if form.body.trim().is_empty() {
        "Must be present".to_string()
    } else if form.body.len() > MAX_COMMENT_LEN {
        format!("Must be shorter than {} characters", MAX_COMMENT_LEN)
    } else {
        let res = state
            .profiles
            .run(CreateComment::from_text(
                submission.id(),
                profile.id(),
                None,
                form.body,
            ))
            .await;

        match res {
            Ok(_) => return Ok(to_submission_page(submission.id())),
            Err(e) => e.to_string(),
        }
    };

    let (nav, current_file, file_num) =
        submission_nav(page.map(|p| p.into_inner()), &submission, &nav_state);

    let comments = CommentNode::from_submission(submission.id(), Some(profile.id()), &state)?;

    let is_self = profile.id() == poster.id();

    let mut comment_input = comment_input(nav_state.dark());
    comment_input.error_opt(Some(error));

    let view = SubmissionView {
        profile: Some(profile),
        submission,
        poster,
        current_file,
        file_num,
        nav,
        is_self,
        comment_input,
        comments,
    };

    crate::rendered(HttpResponse::Ok(), |cursor| {
        crate::templates::submissions::public(cursor, &view, &nav_state)
    })
}

#[derive(Clone, Debug, serde::Deserialize)]
struct UpdateSubmissionForm {
    title: String,
    description: String,
}

async fn update_submission(
    form: web::Form<UpdateSubmissionForm>,
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    use hyaenidae_profiles::apub::actions::UpdateSubmission;
    let form = form.into_inner();

    if profile.id() != submission.profile_id() {
        return Ok(crate::to_404());
    }

    let mut title_error = validate_title(&form.title);
    let description_error = validate_description(&form.description);

    if title_error.is_none() && description_error.is_none() {
        let res = state
            .profiles
            .run(UpdateSubmission::from_text(
                submission.id(),
                form.title.clone(),
                Some(form.description.clone()),
            ))
            .await;

        title_error = match res {
            Ok(_) => return Ok(to_update_page(submission.id())),
            Err(e) => Some(e.to_string()),
        };
    }

    let mut state = SubmissionState::new(submission, nav_state.dark());
    if let Some(error) = title_error {
        state.title_error(error);
    }
    state.description_error(description_error);
    state.title(&form.title);
    state.description(&form.description);

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::update(cursor, &state, &nav_state)
    })
}

async fn add_file(
    req: HttpRequest,
    pl: web::Payload,
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    use hyaenidae_profiles::apub::actions::UpdateSubmission;

    if profile.id() != submission.profile_id() {
        return Ok(crate::to_404());
    }

    let res = state.profiles.upload_image(req, pl.into_inner()).await;

    let error = match res {
        Ok(files) if files.len() == 1 => {
            let res = state
                .profiles
                .run(UpdateSubmission::from_new_file(submission.id(), files[0]))
                .await;

            match res {
                Ok(_) => return Ok(to_update_page(submission.id())),
                Err(e) => e.to_string(),
            }
        }
        Ok(_) => "Invalid number of files".to_owned(),
        Err(e) => e.to_string(),
    };

    let mut state = SubmissionState::new(submission, nav_state.dark());
    state.file_error(error);

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::update(cursor, &state, &nav_state)
    })
}

#[derive(Clone, Debug, serde::Deserialize)]
struct RemoveFileForm {
    file_id: Uuid,
}

async fn remove_file(
    form: web::Form<RemoveFileForm>,
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    use hyaenidae_profiles::apub::actions::UpdateSubmission;

    if profile.id() != submission.profile_id() {
        return Ok(crate::to_404());
    }

    let res = state
        .profiles
        .run(UpdateSubmission::from_removed_file(
            submission.id(),
            form.file_id,
        ))
        .await;

    let error = match res {
        Ok(_) => return Ok(to_update_page(submission.id())),
        Err(e) => e.to_string(),
    };

    let mut state = SubmissionState::new(submission, nav_state.dark());
    state.remove_file_error(form.file_id, error);

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::update(cursor, &state, &nav_state)
    })
}

async fn publish_submission(
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    use hyaenidae_profiles::apub::actions::UpdateSubmission;

    if profile.id() != submission.profile_id() {
        return Ok(crate::to_404());
    }

    let error = if !submission.has_title() {
        "Title must be present".to_owned()
    } else {
        let res = state
            .profiles
            .run(UpdateSubmission::publish_now(submission.id()))
            .await;

        match res {
            Ok(_) => return Ok(to_submission_page(submission.id())),
            Err(e) => e.to_string(),
        }
    };

    let mut state = SubmissionState::new(submission, nav_state.dark());
    state.publish_error(error);

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::update(cursor, &state, &nav_state)
    })
}

async fn delete_submission(
    submission: Submission,
    profile: Profile,
    nav_state: NavState,
    state: web::Data<State>,
) -> Result<HttpResponse, Error> {
    use hyaenidae_profiles::apub::actions::DeleteSubmission;

    if profile.id() != submission.profile_id() {
        return Ok(crate::to_404());
    }

    let res = state
        .profiles
        .run(DeleteSubmission::from_id(submission.id()))
        .await;

    let error = match res {
        Ok(_) => return Ok(crate::to_home()),
        Err(e) => e.to_string(),
    };

    let mut state = SubmissionState::new(submission, nav_state.dark());
    state.delete_error(error);

    crate::rendered(HttpResponse::BadRequest(), |cursor| {
        crate::templates::submissions::update(cursor, &state, &nav_state)
    })
}

async fn route_to_update_page(path: web::Path<Uuid>) -> HttpResponse {
    to_update_page(path.into_inner())
}

fn to_update_page(id: Uuid) -> HttpResponse {
    crate::redirect(&format!("/submissions/{}/update", id))
}

fn to_submission_page(id: Uuid) -> HttpResponse {
    crate::redirect(&format!("/submissions/{}", id))
}

fn file_input(dark: bool) -> FileInput {
    let mut input = FileInput::secondary("images[]", "Select Image");
    input.dark(dark);
    input
}

fn title_input(dark: bool) -> TextInput {
    let mut input = TextInput::new("title");
    input.title("Title").placeholder("Title").dark(dark);
    input
}

fn description_input(dark: bool) -> TextInput {
    let mut input = TextInput::new("description");
    input
        .title("Description")
        .placeholder("Description")
        .textarea()
        .dark(dark);
    input
}
